@@ -11,6 +11,8 @@ from .measure_jackknife import MeasureJackknife
 # import backend method classes used in MeasureIA
 from .measure_w_box import MeasureWBox
 from .measure_m_box import MeasureMultipolesBox
+from .measure_w_box_jk import MeasureWBoxJackknife
+from .measure_m_box_jk import MeasureMBoxJackknife
 from .measure_w_lightcone import MeasureWLightcone
 from .measure_m_lightcone import MeasureMultipolesLightcone
 
