@@ -27,6 +27,33 @@ class MeasureIABase(SimInfo):
 	mu_r_bins : ndarray
 		Bin edges of the mu_r.
 
+	Methods
+	-------
+	calculate_dot_product_arrays()
+		Calculates dot product of elements of two arrays
+	get_ellipticity()
+		Given e and phi, e_+ and e_x components of ellipticity are returned.
+	get_random_pairs()
+		Analytical RR for a (rp,pi) bin.
+	get_volume_spherical_cap()
+		Volume of an (r,mu_r) bin.
+	get_random_pairs_r_mur()
+		Analytical RR for a (r,mu_r) bin.
+	setdiff2D()
+		Compares each row of a1 and a2 and returns the elements that do not overlap.
+	setdiff_omit()
+		For rows in nested list a1, whose index is included in incl_ind, returns elements that do not overlap between
+		the row in a1 and a2.
+	_measure_w_g_i()
+		Measure wgg or wg+ from xi grid provided by MeasureWBox or MeasureWLightcone class methods.
+	_measure_multipoles()
+		Measure multipoles (gg or g+) from xi grid provided by MeasureMultipolesBox or MeasureMultipolesLightcone
+		class methods.
+	_obs_estimator()
+		Combines elements (DD, RR, etc) of xi estimators into xi_gg or xi_g+ for MeasureIALightcone.
+	assign_jackknife_patches()
+		Given positions of multiple samples, defines jackknife patches and returns index of every object in the sample.
+
 	Notes
 	-----
 	Inherits attributes from 'SimInfo', where 'boxsize', 'L_0p5' and 'snap_group' are used in this class.
@@ -158,19 +185,18 @@ class MeasureIABase(SimInfo):
 
 	@staticmethod
 	def calculate_dot_product_arrays(a1, a2):
-		"""Calculates the dot product over 2 2D arrays across axis 1 so that
-		dot_product[i] = np.dot(a1[i],a2[i])
+		"""Calculates the dot product over 2 2D arrays across axis 1 so that dot_product[i] = np.dot(a1[i],a2[i])
 
 		Parameters
 		----------
-		a1 :
+		a1 : ndarray
 			First array
-		a2 :
+		a2 : ndarray
 			Second array
 
 		Returns
 		-------
-		type
+		ndarray
 			Dot product of columns of arrays
 
 		"""
@@ -179,65 +205,6 @@ class MeasureIABase(SimInfo):
 			dot_product += a1[:, i] * a2[:, i]
 		return dot_product
 
-	def measure_3D_orientation_separation_correlation(self, masks=None, dataset_name="All_galaxies"):
-		"""NEEDS MORE EXTENSIVE TESTS
-		Measures the 3D orientation-separation correlation function for given positions and minor axis directions.
-
-		Parameters
-		----------
-		masks :
-			Directory of masks for the data that makes a selection in the data. (Default value = None)
-		dataset_name :
-			Name of the dataset in the hdf5 file specified in output file name. (Default value = "All_galaxies")
-
-		Returns
-		-------
-		type
-			correlation, separation bin means (log) if output file name not specified.
-
-		"""
-		print("WARNING: this method has not been tested and is likely not correct.")
-		exit()
-		if masks == None:
-			positions = self.data["Position"]
-			positions_shape_sample = self.data["Position_shape_sample"]
-			axis_direction_v = self.data["Axis_Direction"]
-			axis_direction_len = np.sqrt(np.sum(axis_direction_v ** 2, axis=1))
-			axis_direction = (axis_direction_v.transpose() / axis_direction_len).transpose()
-		else:
-			positions = self.data["Position"][masks["Position"]]
-			positions_shape_sample = self.data["Position_shape_sample"][masks["Position_shape_sample"]]
-			axis_direction_v = self.data["Axis_Direction"][masks["Axis_Direction"]]
-			axis_direction_len = np.sqrt(np.sum(axis_direction_v ** 2, axis=1))
-			axis_direction = (axis_direction_v.transpose() / axis_direction_len).transpose()
-		n_pairs = [0] * self.num_bins_r
-		inner_product = [0] * self.num_bins_r
-		for n in np.arange(0, len(positions)):
-			separation = positions_shape_sample - positions[n]
-			separation[separation > self.L_0p5] -= self.boxsize
-			separation[separation < -self.L_0p5] += self.boxsize
-			separation_len = np.sqrt(np.sum(separation ** 2, axis=1))
-			separation_dir = (separation.transpose() / separation_len).transpose()
-			inner_product_n = self.calculate_dot_product_arrays(separation_dir, axis_direction) ** 2
-			for i in np.arange(0, self.num_bins_r):
-				lower_limit_mask = separation_len > self.r_bins[i]
-				upper_limit_mask = separation_len < self.r_bins[i + 1]
-				mask = lower_limit_mask * upper_limit_mask
-				n_pairs[i] += sum(mask)
-				inner_product[i] += sum(inner_product_n[mask])
-		correlation = np.array(inner_product) / np.array(n_pairs) - 1.0 / 3
-		dsep = (self.r_bins[:-1] - self.r_bins[1:]) / 2.0
-		separation_bins = self.r_bins[:-1] + abs(dsep)
-
-		if self.output_file_name != None:
-			output_file = h5py.File(self.output_file_name, "a")
-			group = create_group_hdf5(output_file, f"{self.snap_group}/3D_correlations")
-			write_dataset_hdf5(group, dataset_name, data=np.array([separation_bins, correlation]).transpose())
-			output_file.close()
-			return
-		else:
-			return correlation, separation_bins
-
 	@staticmethod
 	def get_ellipticity(e, phi):
 		"""Calculates the radial and tangential components of the ellipticity, given the size of the ellipticty vector
@@ -245,14 +212,14 @@ class MeasureIABase(SimInfo):
 
 		Parameters
 		----------
-		e :
+		e : ndarray
 			size of the ellipticity vector
-		phi :
+		phi : ndarray
 			angle between semimajor/semiminor axis and separation vector
 
 		Returns
 		-------
-		type
+		ndarray
 			e_+ and e_x
 
 		"""
@@ -266,26 +233,27 @@ class MeasureIABase(SimInfo):
 
 		Parameters
 		----------
-		rp_max :
-			upper bound of projected separation vector bin
-		rp_min :
-			lower bound of projected separation vector bin
-		pi_max :
-			upper bound of line of sight vector bin
-		pi_min :
-			lower bound of line of sight vector bin
-		L3 :
-			volume of the simulation box
-		corrtype :
+		rp_max : float
+			Upper bound of projected separation vector bin
+		rp_min : float
+			Lower bound of projected separation vector bin
+		pi_max : float
+			Upper bound of line of sight vector bin
+		pi_min : float
+			Lower bound of line of sight vector bin
+		L3 : float or int
+			Volume of the simulation box
+		corrtype : str
 			Correlation type, auto or cross. RR for auto is RR_cross/2.
-		Num_position :
-
-		Num_shape :
+		Num_position : int
+			Number of objects in the position sample.
+		Num_shape : int
+			Number of objects in the shape sample.
 
 
 		Returns
 		-------
-		type
+		float
 			number of pairs in r_p, pi bin
 
 		"""
@@ -309,45 +277,45 @@ class MeasureIABase(SimInfo):
 
 		Parameters
 		----------
-		mur :
+		mur : float
 			cos(theta), where theta is the polar angle between the apex and disk of the cap.
-		r :
+		r : float
 			radius
 
 		Returns
 		-------
-		type
+		float
 			Volume of the spherical cap.
 
 		"""
 		return np.pi / 3.0 * r ** 3 * (2 + mur) * (1 - mur) ** 2
 
 	def get_random_pairs_r_mur(self, r_max, r_min, mur_max, mur_min, L3, corrtype, Num_position, Num_shape):
-		"""Retruns analytical value of the number of pairs expected in an r_p, pi bin for a random uniform distribution.
-		(Singh et al. 2023)
+		"""Returns analytical value of the number of pairs expected in an r, mu_r bin for a random uniform distribution.
 
 		Parameters
 		----------
-		r_max :
-			upper bound of projected separation vector bin
-		r_min :
-			lower bound of projected separation vector bin
-		mur_max :
-			upper bound of mu_r bin
-		mur_min :
-			lower bound of mu_r bin
-		L3 :
-			volume of the simulation box
-		corrtype :
+		r_max : float
+			Upper bound of separation vector bin
+		r_min : float
+			Lower bound of separation vector bin
+		mur_max : float
+			Upper bound of mu_r bin
+		mur_min : float
+			Lower bound of mu_r bin
+		L3 : float
+			Volume of the simulation box
+		corrtype : str
 			Correlation type, auto or cross. RR for auto is RR_cross/2.
-		Num_position :
-
-		Num_shape :
+		Num_position : int
+			Number of objects in the position sample.
+		Num_shape : int
+			Number of objects in the shape sample.
 
 
 		Returns
 		-------
-		type
+		float
 			number of pairs in r, mu_r bin
 
 		"""
@@ -384,19 +352,21 @@ class MeasureIABase(SimInfo):
 
 	@staticmethod
 	def setdiff2D(a1, a2):
-		"""
+		"""Compares each row of a1 and a2 and returns the elements that do not overlap
 
 		Parameters
 		----------
-		a1 :
-
-		a2 :
-
+		a1 : nested list
+			List containing lists of elements to compare to a2
+		a2 : nested list
+			List containing lists of elements to compare to a1
 
 		Returns
 		-------
-
+		nested list
+			For each row, the not-overlapping elements between a1 and a2
 		"""
+		assert len(a1) == len(a2), "Lengths of lists where each row is to be compared, are not the same."
 		diff = []
 		for i in np.arange(0, len(a1)):
 			setdiff = np.setdiff1d(a1[i], a2[i])
@@ -406,20 +376,22 @@ class MeasureIABase(SimInfo):
 
 	@staticmethod
 	def setdiff_omit(a1, a2, incl_ind):
-		"""
+		"""For rows in nested list a1, whose index is included in incl_ind, returns elements that do not overlap between
+		the row in a1 and a2.
 
 		Parameters
 		----------
-		a1 :
-
-		a2 :
-
-		incl_ind :
-
+		a1 : nested list
+			List of lists or arrays where indicated rows need to be compared to a2
+		a2 : list or array
+			Elements to be compared to the row in a1 [and not included in return values].
+		incl_ind : list or array
+			Indices of rows in a1 to be compared to a2.
 
 		Returns
 		-------
-
+		nested list
+			For each included row in a1, the not-overlapping elements between a1 and a2
 		"""
 		diff = []
 		for i in np.arange(0, len(a1)):
@@ -429,113 +401,136 @@ class MeasureIABase(SimInfo):
 				del setdiff
 		return diff
 
-
-
-	def measure_projected_correlation_save_pairs(self, output_file_pairs="", masks=None, dataset_name="All_galaxies",
-												 print_num=True):
-		"""Measures the projected correlation function (xi_g_plus, xi_gg) for given coordinates of the position and shape sample
-		(Position, Position_shape_sample), the projected axis direction (Axis_Direction), the ratio between projected
-		axes, q=b/a (q) and the index of the direction of the line of sight (LOS=2 for z axis).
-		Positions are assumed to be given in cMpc/h.
+	def _get_jackknife_region_indices(self, masks, L_subboxes):
+		"""
+		Split the box in L_subboxes^3 subboxes and return indices of which subbox objects are in for position and
+		shape sample.
 
 		Parameters
 		----------
-		masks :
-			the masks for the data to select only part of the data (Default value = None)
-		dataset_name :
-			the dataset name given in the hdf5 file. (Default value = "All_galaxies")
-		return_output :
-			Output is returned if True, saved to file if False.
-		output_file_pairs :
-			 (Default value = "")
-		print_num :
-			 (Default value = True)
+		masks: dict or NoneType
+			Input in methods in MeasureIABox that masks the input data dictionary.
+		L_subboxes: int
+			Number of subboxes on one side of the box. L_subboxes^3 is the total number of jackknife realisations.
 
 		Returns
 		-------
-		type
-			xi_g_plus, xi_gg, separation_bins, pi_bins if no output file is specified
+		ndarrays
+			indices of jackknife region of position sample and indices of jackknife region of shape sample
 
 		"""
-
 		if masks == None:
 			positions = self.data["Position"]
 			positions_shape_sample = self.data["Position_shape_sample"]
-			axis_direction_v = self.data["Axis_Direction"]
-			axis_direction_len = np.sqrt(np.sum(axis_direction_v ** 2, axis=1))
-			axis_direction = (axis_direction_v.transpose() / axis_direction_len).transpose()
-			q = self.data["q"]
 		else:
 			positions = self.data["Position"][masks["Position"]]
 			positions_shape_sample = self.data["Position_shape_sample"][masks["Position_shape_sample"]]
-			axis_direction_v = self.data["Axis_Direction"][masks["Axis_Direction"]]
-			axis_direction_len = np.sqrt(np.sum(axis_direction_v ** 2, axis=1))
-			axis_direction = (axis_direction_v.transpose() / axis_direction_len).transpose()
-			q = self.data["q"][masks["q"]]
-		Num_position = len(positions)
-		Num_shape = len(positions_shape_sample)
-		if print_num:
-			print(
-				f"There are {Num_shape} galaxies in the shape sample and {Num_position} galaxies in the position sample.")
+		L_sub = self.L_0p5 * 2.0 / L_subboxes
+		jackknife_region_indices_pos = np.zeros(len(positions))
+		jackknife_region_indices_shape = np.zeros(len(positions_shape_sample))
+		num_box = 0
+		for i in np.arange(0, L_subboxes):
+			for j in np.arange(0, L_subboxes):
+				for k in np.arange(0, L_subboxes):
+					x_bounds = [i * L_sub, (i + 1) * L_sub]
+					y_bounds = [j * L_sub, (j + 1) * L_sub]
+					z_bounds = [k * L_sub, (k + 1) * L_sub]
+					x_mask = (positions[:, 0] > x_bounds[0]) * (positions[:, 0] < x_bounds[1])
+					y_mask = (positions[:, 1] > y_bounds[0]) * (positions[:, 1] < y_bounds[1])
+					z_mask = (positions[:, 2] > z_bounds[0]) * (positions[:, 2] < z_bounds[1])
+					x_mask_shape = (positions_shape_sample[:, 0] > x_bounds[0]) * (
+								positions_shape_sample[:, 0] < x_bounds[1])
+					y_mask_shape = (positions_shape_sample[:, 1] > y_bounds[0]) * (
+								positions_shape_sample[:, 1] < y_bounds[1])
+					z_mask_shape = (positions_shape_sample[:, 2] > z_bounds[0]) * (
+								positions_shape_sample[:, 2] < z_bounds[1])
+					mask_position = x_mask * y_mask * z_mask  # mask that is True for all positions in the subbox
+					mask_shape = x_mask_shape * y_mask_shape * z_mask_shape  # mask that is True for all positions not in the subbox
+					jackknife_region_indices_pos[mask_position] = num_box
+					jackknife_region_indices_shape[mask_shape] = num_box
+					num_box += 1
+		return np.array(jackknife_region_indices_pos, dtype=int), np.array(jackknife_region_indices_shape, dtype=int)
 
-		LOS_ind = self.data["LOS"]  # eg 2 for z axis
-		not_LOS = np.array([0, 1, 2])[np.isin([0, 1, 2], LOS_ind, invert=True)]  # eg 0,1 for x&y
-		e = (1 - q ** 2) / (1 + q ** 2)  # size of ellipticity
-		del q
-		R = 1 - np.mean(e ** 2) / 2.0  # responsitivity factor
-		output_file_pairs = h5py.File(output_file_pairs, "a")
-		group = create_group_hdf5(output_file_pairs, "w_g_plus")
-
-		indices_shape = np.arange(0, len(positions_shape_sample))
-		for n in np.arange(0, len(positions)):
-			# for Splus_D (calculate ellipticities around position sample)
-			separation = positions_shape_sample - positions[n]
-			if self.periodicity:
-				separation[separation > self.L_0p5] -= self.boxsize  # account for periodicity of box
-				separation[separation < -self.L_0p5] += self.boxsize
-			projected_sep = separation[:, not_LOS]
-			LOS = separation[:, LOS_ind]
-			del separation
-			separation_len = np.sqrt(np.sum(projected_sep ** 2, axis=1))
-			separation_dir = (projected_sep.transpose() / separation_len).transpose()  # normalisation of rp
-			del projected_sep
-			phi = np.arccos(self.calculate_dot_product_arrays(separation_dir, axis_direction))  # [0,pi]
-			del separation_dir
-			e_plus, e_cross = self.get_ellipticity(e, phi)
-			del phi
-			e_plus[np.isnan(e_plus)] = 0.0
-			e_cross[np.isnan(e_cross)] = 0.0
-
-			write_data = (np.array(
-				[[n] * len(indices_shape), indices_shape, separation_len, LOS, e_plus / (2 * R)]).transpose())
-			# np.array(
-			# [[n] * len(ind_r), indices_shape[mask], ind_r, ind_pi, e_plus[mask] / (2 * R)]).transpose())
-			if n == 0:
-				group.create_dataset(dataset_name, data=write_data, maxshape=(None, 5), chunks=True)
-			else:
-				group[dataset_name].resize((group[dataset_name].shape[0] + write_data.shape[0]), axis=0)
-				group[dataset_name][-write_data.shape[0]:] = write_data
-
-		output_file_pairs.close()
-		return
-
-	def _measure_w_g_i(self, corr_type="both", dataset_name="All_galaxies", return_output=False, jk_group_name=""):
-		"""Measures w_gi for a given xi_gi dataset that has been calculated with the measure projected correlation
-		method. Sums over pi values. Stores [rp, w_gi]. i can be + or g
+	def _combine_jackknife_information(self, dataset_name, jk_group_name, corr_group, num_box, return_output=False):
+		"""
+		Combine jackknife realisations into a covariance matrix.
 
 		Parameters
 		----------
-		dataset_name :
-			Name of xi_gi dataset and name given to w_gi dataset when stored. (Default value = "All_galaxies")
-		return_output :
-			Output is returned if True, saved to file if False. (Default value = False)
-		corr_type :
-			Type of correlation function. Choose from [g+,gg,both]. (Default value = "both")
-		jk_group_name :
-			 (Default value = "")
+		dataset_name: str
+			Name of the dataset in the output file.
+		jk_group_name: str
+			Name of the subgroup in the output file where the jackknife realisations are saved.
+		corr_group: list of str
+			Name of the subgroups in the output file denoting the correlation (e.g. w_g_plus, multipoles_gg etc).
+		num_box: int
+			Number of jackknife realisations.
+		return_output: bool, optional
+			When True, returns output, otherwise saves to output file.
 
 		Returns
 		-------
+		list of ndarrays
+			list of covariances for each entry in corr_group and list of standard deviations for each entry in corr_group
+
+		"""
+		covs, stds = [], []
+		for d in np.arange(0, len(corr_group)):
+			data_file = h5py.File(self.output_file_name, "a")
+			group_multipoles = data_file[f"{self.snap_group}/{corr_group[d]}/{jk_group_name}/"]
+			# calculating mean of the datavectors
+			mean_multipoles = np.zeros(self.num_bins_r)
+			for b in np.arange(0, num_box):
+				mean_multipoles += group_multipoles[dataset_name + "_" + str(b)][:]
+			mean_multipoles /= num_box
+
+			# calculation the covariance matrix (multipoles) and the standard deviation (sqrt of diag of cov)
+			cov = np.zeros((self.num_bins_r, self.num_bins_r))
+			std = np.zeros(self.num_bins_r)
+			for b in np.arange(0, num_box):
+				std += (group_multipoles[dataset_name + "_" + str(b)][:] - mean_multipoles) ** 2
+				for i in np.arange(self.num_bins_r):
+					cov[:, i] += (group_multipoles[dataset_name + "_" + str(b)][:] - mean_multipoles) * (
+							group_multipoles[dataset_name + "_" + str(b)][i] - mean_multipoles[i]
+					)
+			std *= (num_box - 1) / num_box  # see Singh 2023
+			std = np.sqrt(std)  # size of errorbars
+			cov *= (num_box - 1) / num_box  # cov not sqrt so to get std, sqrt of diag would need to be taken
+			data_file.close()
+			if return_output:
+				covs.append(cov)
+				stds.append(std)
+			else:
+				output_file = h5py.File(self.output_file_name, "a")
+				group_multipoles = create_group_hdf5(output_file, f"{self.snap_group}/" + corr_group[d])
+				write_dataset_hdf5(group_multipoles, dataset_name + "_mean_" + str(num_box), data=mean_multipoles)
+				write_dataset_hdf5(group_multipoles, dataset_name + "_jackknife_" + str(num_box), data=std)
+				write_dataset_hdf5(group_multipoles, dataset_name + "_jackknife_cov_" + str(num_box), data=cov)
+				output_file.close()
+		if return_output:
+			return covs, stds
+		else:
+			return
+
+	def _measure_w_g_i(self, dataset_name, corr_type="both", return_output=False, jk_group_name=""):
+		"""Measures w_gg or w_g+ for a given xi_gi dataset that has been calculated with the _measure_xi_rp_pi_sims
+		methods. Integrates over pi bins via sum * dpi. Stores rp, and w_gg or w_g+.
+
+		Parameters
+		----------
+		dataset_name : str
+			Name of xi_gg or xi_g+ dataset and name given to w_gg or w_g+ dataset when stored.
+		return_output : bool, optional
+			Output is returned if True, saved to file if False. Default value = False
+		corr_type : str, optional
+			Type of correlation function. Choose from [g+,gg,both]. Default value = "both"
+		jk_group_name : str, optional
+			Name of subgroup in hdf5 file where jackknife realisations are stored. Default value = ""
+
+		Returns
+		-------
+		ndarray
+			[rp, wgg] or [rp, wg+] if return_output is True
 
 		"""
 		if corr_type == "both":
@@ -579,25 +574,25 @@ class MeasureIABase(SimInfo):
 				correlation_data_file.close()
 		return
 
-	def _measure_multipoles(self, corr_type="both", dataset_name="All_galaxies", return_output=False, jk_group_name=""):
-		"""Measures multipoles for a given xi_g+ calculated by measure projected correlation.
-		The data assumes xi_g+ to be measured in bins of rp and pi. It measures mu_r and r
-		and saves the multipoles in the (r,mu_r) space. Should be binned into r bins.
+	def _measure_multipoles(self, dataset_name, corr_type="both", return_output=False, jk_group_name=""):
+		"""Measures multipoles for a given xi_g+ or xi_gg measured by _measure_xi_r_pi_sims methods.
+		The data assumes xi_g+ and xi_gg to be measured in bins of r and mu_r.
 
 		Parameters
 		----------
-		corr_type :
-			Default value of g+, ensuring correct dataset and sab and l to be 2.
-		dataset_name :
-			Name of the dataset of xi_g+ and multipoles. (Default value = "All_galaxies")
-		return_output :
-			Output is returned if True, saved to file if False. (Default value = False)
-		jk_group_name :
-			 (Default value = "")
+		dataset_name : str
+			Name of xi_gg or xi_g+ dataset and name given to multipoles dataset when stored.
+		corr_type : str, optional
+			Type of correlation function. Choose from [g+,gg,both]. Default value = "both"
+		return_output : bool, optional
+			Output is returned if True, saved to file if False. Default value = False.
+		jk_group_name : str, optional
+			Name of subgroup in hdf5 file where jackknife realisations are stored. Default value = ""
 
 		Returns
 		-------
-
+		ndarray
+			[r, multipoles_gg] or [r, multipoles_g+] if return_output is True
 		"""
 		correlation_data_file = h5py.File(self.output_file_name, "a")
 		if corr_type == "g+":  # todo: expand to include ++ option
@@ -673,23 +668,24 @@ class MeasureIABase(SimInfo):
 		return
 
 	def _obs_estimator(self, corr_type, IA_estimator, dataset_name, dataset_name_randoms, num_samples,
-					   jk_group_name=""):
-		"""Reads various components of xi and combines into correct estimator for cluster or galaxy observational alignments
+					   jk_group_name="", jk_group_name_randoms=""):
+		"""Reads various components of xi and combines into correct estimator for cluster or galaxy
+		lightcone alignment correlations. It then writes the xi_gg or xi_g+ in the correct place in the output file.
 
 		Parameters
 		----------
-		corr_type :
-			w or multipoles
-		IA_estimator :
-			clusters or galaxies
-		dataset_name :
+		corr_type : list of 2 str elements
+			First element: ['gg', 'g+', 'both'], second: 'w' or 'multipoles'
+		IA_estimator : str
+			Chooser from 'clusters' or 'galaxies' for different estimator definition.
+		dataset_name : str
 			Name of the dataset
-		dataset_name_randoms :
+		dataset_name_randoms : str
 			Name of the dataset for data with randoms as positions
-		num_samples :
-
-		jk_group_name :
-			 (Default value = "")
+		num_samples : dict
+			Dictionary of samples sizes for position, shape and random samples. Keywords: D, S, R_D, R_S
+		jk_group_name : str
+			Name of subgroup in hdf5 file where jackknife realisations are stored. Default value = ""
 
 		Returns
 		-------
@@ -698,17 +694,20 @@ class MeasureIABase(SimInfo):
 		output_file = h5py.File(self.output_file_name, "a")
 		if corr_type[0] == "g+" or corr_type[0] == "both":
 			group_gp = output_file[
-				f"{self.snap_group}/{corr_type[1]}/xi_g_plus/{jk_group_name}"]  # /w/xi_g_plus/
+				f"{self.snap_group}/{corr_type[1]}/xi_g_plus/{jk_group_name}"]
+			group_gp_r = output_file[
+				f"{self.snap_group}/{corr_type[1]}/xi_g_plus/{jk_group_name_randoms}"]
 			SpD = group_gp[f"{dataset_name}_SplusD"][:]
-			SpR = group_gp[f"{dataset_name_randoms}_SplusD"][:]
+			SpR = group_gp_r[f"{dataset_name_randoms}_SplusD"][:]
 		group_gg = output_file[f"{self.snap_group}/{corr_type[1]}/xi_gg/{jk_group_name}"]
+		group_gg_r = output_file[f"{self.snap_group}/{corr_type[1]}/xi_gg/{jk_group_name_randoms}"]
 		DD = group_gg[f"{dataset_name}_DD"][:]
 
 		if IA_estimator == "clusters":
 			if corr_type[0] == "gg":
 				SR = group_gg[f"{dataset_name}_SR"][:]
 			else:
-				SR = group_gg[f"{dataset_name_randoms}_DD"][:]
+				SR = group_gg_r[f"{dataset_name_randoms}_DD"][:]
 			SR *= num_samples["D"] / num_samples["R_D"]
 			if corr_type[0] == "g+" or corr_type[0] == "both":
 				SpR *= num_samples["D"] / num_samples["R_D"]
@@ -733,7 +732,7 @@ class MeasureIABase(SimInfo):
 				if corr_type[0] == "gg":
 					SR = group_gg[f"{dataset_name}_SR"][:]
 				else:
-					SR = group_gg[f"{dataset_name_randoms}_DD"][:]
+					SR = group_gg_r[f"{dataset_name_randoms}_DD"][:]
 				RD *= num_samples["S"] / num_samples["R_S"]
 				SR *= num_samples["D"] / num_samples["R_D"]
 				correlation_gg = (DD - RD - SR) / RR - 1
@@ -749,17 +748,20 @@ class MeasureIABase(SimInfo):
 
 		Parameters
 		----------
-		data :
-			directory containing position and shape sample data
-		randoms_data :
-			directory containing position and shape sample data of randoms
-		num_jk :
-			number of jackknife patches
+		data : dict
+			Dictionary containing position and shape sample data. Keywords: "RA", "DEC", "RA_shape_sample",
+			"DEC_shape_sample"
+		randoms_data : dict
+			Dictionary containing position and shape sample data of randoms. Keywords: "RA", "DEC", "RA_shape_sample",
+			"DEC_shape_sample"
+		num_jk : int
+			Number of jackknife patches
 
 		Returns
 		-------
-		type
-			directory with patch numbers for each sample
+		dict
+			Dictionary with patch numbers for each sample. Keywords: 'position', 'shape', 'randoms_position',
+			'randoms_shape'
 
 		"""
 
@@ -799,43 +801,6 @@ class MeasureIABase(SimInfo):
 
 		return jk_patches
 
-	def measure_misalignment_angle(self, vector1_name, vector2_name, normalise=False):
-		"""NOT TESTED
-		Calculate the misalignment angle between two given vectors. Assumes the vectors to be normalised unless
-		otherwise specified.
-
-		Parameters
-		----------
-		vector1_name :
-			Name in data of the first vector.
-		vector2_name :
-			Name in data of the second vector
-		normalise :
-			If True, the vectors are divided by their length. Default is False.
-
-		Returns
-		-------
-		type
-			the misalignment angle, unless an output file name is given.
-
-		"""
-
-		eigen_vector1 = self.data[vector1_name]
-		eigen_vector2 = self.data[vector2_name]
-		if normalise:
-			eigen_vector1 = (eigen_vector1.transpose() / np.sqrt(np.sum(eigen_vector1 ** 2, axis=1))).transpose()
-			eigen_vector2 = (eigen_vector2.transpose() / np.sqrt(np.sum(eigen_vector2 ** 2, axis=1))).transpose()
-		misalignment_angle = np.arccos(self.calculate_dot_product_arrays(eigen_vector1, eigen_vector2))
-
-		if self.output_file_name != None:
-			output_file = h5py.File(self.output_file_name, "a")
-			group = create_group_hdf5(output_file, f"{self.snap_group}/Misalignment_angels")
-			write_dataset_hdf5(group, vector1_name + "_" + vector2_name, data=misalignment_angle)
-			output_file.close()
-		else:
-			return misalignment_angle
-		return
-
 
 if __name__ == "__main__":
 	pass
