@@ -1,19 +1,32 @@
 import numpy as np
 import h5py
-import pickle
-from pathos.multiprocessing import ProcessingPool
+import os
+from multiprocessing import Pool
 from scipy.spatial import KDTree
 from .write_data import write_dataset_hdf5, create_group_hdf5
 from .measure_IA_base import MeasureIABase
+from .read_data import ReadData
 from astropy.cosmology import LambdaCDM
 
 cosmo = LambdaCDM(H0=69.6, Om0=0.286, Ode0=0.714)
 KPC_TO_KM = 3.086e16  # 1 kpc is 3.086e16 km
 
 
-class MeasureWBox(MeasureIABase):
-	"""Class that contains all methods for the measurements of xi_gg and xi_g+ for w_gg and w_g+ with carthesian
-	simulation data.
+class MeasureWBox(MeasureIABase, ReadData):
+	r"""Class that contains all methods for the measurements of $\xi_{gg}$ and $\xi_{g+}$ for $w_{gg}$ and $w_{g+}$
+	 with Cartesian simulation data.
+
+	Methods
+	-------
+	_measure_xi_rp_pi_box_brute()
+		Measure $\xi_{gg}$ and $\xi_{g+}$ in (rp, pi) grid binning in a periodic box using 1 CPU.
+	_measure_xi_rp_pi_box_tree()
+		Measure $\xi_{gg}$ and $\xi_{g+}$ in (rp, pi) grid binning in a periodic box using 1 CPU and KDTree for extra speed.
+	_measure_xi_rp_pi_box_batch()
+		Measure $\xi_{gg}$ and $\xi_{g+}$ in (rp, pi) grid binning in a periodic box using 1 CPU for a batch of indices.
+		Support function of _measure_xi_rp_pi_box_multiprocessing().
+	_measure_xi_rp_pi_box_multiprocessing()
+		Measure $\xi_{gg}$ and $\xi_{g+}$ in (rp, pi) grid binning in a periodic box using >1 CPUs.
 
 	Notes
 	-----
@@ -49,32 +62,30 @@ class MeasureWBox(MeasureIABase):
 						 pi_max, boxsize, periodicity)
 		return
 
-	def _measure_xi_rp_pi_sims_brute(self, masks=None, dataset_name="All_galaxies", return_output=False,
-									 print_num=True,
-									 jk_group_name=""):
-		"""Measures the projected correlation function (xi_g_plus, xi_gg) for given coordinates of the position and shape sample
-		(Position, Position_shape_sample), the projected axis direction (Axis_Direction), the ratio between projected
-		axes, q=b/a (q) and the index of the direction of the line of sight (LOS=2 for z axis).
-		Positions are assumed to be given in cMpc/h.
+	def _measure_xi_rp_pi_box_brute(self, dataset_name, masks=None, return_output=False,
+									jk_group_name="", ellipticity='distortion'):
+		r"""Measures the projected correlation functions, $\xi_{gg}$ and $\xi_{g+}$, in (rp, pi) bins for an object
+		created with MeasureIABox. Uses 1 CPU.
 
 		Parameters
 		----------
-		masks :
-		    the masks for the data to select only part of the data (Default value = None)
-		dataset_name :
-		    the dataset name given in the hdf5 file. (Default value = "All_galaxies")
-		return_output :
-		    Output is returned if True, saved to file if False. (Default value = False)
-		print_num :
-		     (Default value = True)
-		jk_group_name :
-		     (Default value = "")
+		dataset_name : str
+			Name of the dataset in the output file.
+		masks : dict or NoneType, optional
+			Dictionary with masks for the data to select only part of the data. Uses same keywords as data dictionary.
+			Default value = None.
+		return_output : bool, optional
+			If True, the output will be returned instead of written to a file. Default value is False.
+		jk_group_name : str, optional
+			Group in output file (hdf5) where jackknife realisations are stored. Default value is "".
+		ellipticity : str, optional
+			Definition of ellipticity. Choose from 'distortion', defined as (1-q^2)/(1+q^2), or 'ellipticity', defined
+			 as (1-q)/(1+q). Default is 'distortion'.
 
 		Returns
 		-------
-		type
-		    xi_g_plus, xi_gg, separation_bins, pi_bins if no output file is specified
-
+		ndarrays
+			$\xi_{gg}$ and $\xi_{g+}$, r_p bins, pi bins, S+D, DD, RR (if no output file is specified)
 		"""
 
 		if masks == None:
@@ -107,13 +118,17 @@ class MeasureWBox(MeasureIABase):
 			weight_shape = self.data["weight_shape_sample"][masks["weight_shape_sample"]]
 		Num_position = len(positions)
 		Num_shape = len(positions_shape_sample)
-		if print_num:
-			print(
-				f"There are {Num_shape} galaxies in the shape sample and {Num_position} galaxies in the position sample.")
+		print(
+			f"There are {Num_shape} galaxies in the shape sample and {Num_position} galaxies in the position sample.")
 
 		LOS_ind = self.data["LOS"]  # eg 2 for z axis
 		not_LOS = np.array([0, 1, 2])[np.isin([0, 1, 2], LOS_ind, invert=True)]  # eg 0,1 for x&y
-		e = (1 - q ** 2) / (1 + q ** 2)  # size of ellipticity
+		if ellipticity == 'distortion':
+			e = (1 - q ** 2) / (1 + q ** 2)  # size of ellipticity
+		elif ellipticity == 'ellipticity':
+			e = (1 - q) / (1 + q)
+		else:
+			raise ValueError("Invalid value for ellipticity. Choose 'distortion' or 'ellipticity'.")
 		del q
 		R = sum(weight_shape * (1 - e ** 2 / 2.0)) / sum(weight_shape)
 		# R = 1 - np.mean(e ** 2) / 2.0  # responsitivity factor
@@ -125,7 +140,6 @@ class MeasureWBox(MeasureIABase):
 		Scross_D = np.array([[0.0] * self.num_bins_pi] * self.num_bins_r)
 		RR_g_plus = np.array([[0.0] * self.num_bins_pi] * self.num_bins_r)
 		RR_gg = np.array([[0.0] * self.num_bins_pi] * self.num_bins_r)
-		variance = np.array([[0.0] * self.num_bins_pi] * self.num_bins_r)
 
 		for n in np.arange(0, len(positions)):
 			# for Splus_D (calculate ellipticities around position sample)
@@ -160,9 +174,12 @@ class MeasureWBox(MeasureIABase):
 			)  # need length of LOS, so only positive values
 			del LOS
 			ind_pi = np.array(ind_pi, dtype=int)
+			if np.any(ind_pi == self.num_bins_pi):
+				ind_pi[ind_pi >= self.num_bins_pi] -= 1
+			if np.any(ind_r == self.num_bins_r):
+				ind_r[ind_r >= self.num_bins_r] -= 1
 			np.add.at(Splus_D, (ind_r, ind_pi), (weight[n] * weight_shape[mask] * e_plus[mask]) / (2 * R))
 			np.add.at(Scross_D, (ind_r, ind_pi), (weight[n] * weight_shape[mask] * e_cross[mask]) / (2 * R))
-			np.add.at(variance, (ind_r, ind_pi), ((weight[n] * weight_shape[mask] * e_plus[mask]) / (2 * R)) ** 2)
 			del e_plus, e_cross
 			np.add.at(DD, (ind_r, ind_pi), weight[n] * weight_shape[mask])
 
@@ -183,7 +200,6 @@ class MeasureWBox(MeasureIABase):
 					Num_position, Num_shape)
 		correlation = Splus_D / RR_g_plus  # (Splus_D - Splus_R) / RR_g_plus
 		xi_g_cross = Scross_D / RR_g_plus  # (Scross_D - Scross_R) / RR_g_plus
-		sigsq = variance / RR_g_plus ** 2
 		dsep = (self.r_bins[1:] - self.r_bins[:-1]) / 2.0
 		separation_bins = self.r_bins[:-1] + abs(dsep)  # middle of bins
 		dpi = (self.pi_bins[1:] - self.pi_bins[:-1]) / 2.0
@@ -195,21 +211,18 @@ class MeasureWBox(MeasureIABase):
 			write_dataset_hdf5(group, dataset_name, data=correlation)
 			write_dataset_hdf5(group, dataset_name + "_SplusD", data=Splus_D)
 			write_dataset_hdf5(group, dataset_name + "_RR_g_plus", data=RR_g_plus)
-			write_dataset_hdf5(group, dataset_name + "_sigmasq", data=sigsq)
 			write_dataset_hdf5(group, dataset_name + "_rp", data=separation_bins)
 			write_dataset_hdf5(group, dataset_name + "_pi", data=pi_bins)
 			group = create_group_hdf5(output_file, f"{self.snap_group}/w/xi_g_cross/{jk_group_name}")
 			write_dataset_hdf5(group, dataset_name + "_ScrossD", data=Scross_D)
 			write_dataset_hdf5(group, dataset_name, data=xi_g_cross)
 			write_dataset_hdf5(group, dataset_name + "_RR_g_cross", data=RR_g_plus)
-			write_dataset_hdf5(group, dataset_name + "_sigmasq", data=sigsq)
 			write_dataset_hdf5(group, dataset_name + "_rp", data=separation_bins)
 			write_dataset_hdf5(group, dataset_name + "_pi", data=pi_bins)
 			group = create_group_hdf5(output_file, f"{self.snap_group}/w/xi_gg/{jk_group_name}")
 			write_dataset_hdf5(group, dataset_name, data=(DD / RR_gg) - 1)
 			write_dataset_hdf5(group, dataset_name + "_DD", data=DD)
 			write_dataset_hdf5(group, dataset_name + "_RR_gg", data=RR_gg)
-			write_dataset_hdf5(group, dataset_name + "_sigmasq", data=sigsq)
 			write_dataset_hdf5(group, dataset_name + "_rp", data=separation_bins)
 			write_dataset_hdf5(group, dataset_name + "_pi", data=pi_bins)
 			output_file.close()
@@ -217,40 +230,30 @@ class MeasureWBox(MeasureIABase):
 		else:
 			return correlation, (DD / RR_gg) - 1, separation_bins, pi_bins, Splus_D, DD, RR_g_plus
 
-	def _measure_xi_rp_pi_sims_tree(self, tree_input=None, masks=None, dataset_name="All_galaxies",
-									return_output=False, print_num=True, dataset_name_tree=None, save_tree=False,
-									file_tree_path=None, jk_group_name=""):
-		"""Measures the projected correlation function (xi_g_plus, xi_gg) for given coordinates of the position and shape sample
-		(Position, Position_shape_sample), the projected axis direction (Axis_Direction), the ratio between projected
-		axes, q=b/a (q) and the index of the direction of the line of sight (LOS=2 for z axis).
-		Positions are assumed to be given in cMpc/h.
+	def _measure_xi_rp_pi_box_tree(self, dataset_name, masks=None,
+								   return_output=False, jk_group_name="", ellipticity='distortion'):
+		r"""Measures the projected correlation functions, $\xi_{gg}$ and $\xi_{g+}$, in (rp, pi) bins for an object
+		created with MeasureIABox. Uses 1 CPU. Uses KDTree for speedup.
 
 		Parameters
 		----------
-		masks :
-		    the masks for the data to select only part of the data (Default value = None)
-		dataset_name :
-		    the dataset name given in the hdf5 file. (Default value = "All_galaxies")
-		return_output :
-		    Output is returned if True, saved to file if False. (Default value = False)
-		tree_input :
-		     (Default value = None)
-		print_num :
-		     (Default value = True)
-		dataset_name_tree :
-		     (Default value = None)
-		save_tree :
-		     (Default value = False)
-		file_tree_path :
-		     (Default value = None)
-		jk_group_name :
-		     (Default value = "")
+		dataset_name : str
+			Name of the dataset in the output file.
+		masks : dict or NoneType, optional
+			Dictionary with masks for the data to select only part of the data. Uses same keywords as data dictionary.
+			Default value = None.
+		return_output : bool, optional
+			If True, the output will be returned instead of written to a file. Default value is False.
+		jk_group_name : str, optional
+			Group in output file (hdf5) where jackknife realisations are stored. Default value is "".
+		ellipticity : str, optional
+			Definition of ellipticity. Choose from 'distortion', defined as (1-q^2)/(1+q^2), or 'ellipticity', defined
+			 as (1-q)/(1+q). Default is 'distortion'.
 
 		Returns
 		-------
-		type
-		    xi_g_plus, xi_gg, separation_bins, pi_bins if no output file is specified
-
+		ndarrays
+			$\xi_{gg}$ and $\xi_{g+}$, r_p bins, pi bins, S+D, DD, RR (if no output file is specified)
 		"""
 
 		if masks == None:
@@ -286,7 +289,12 @@ class MeasureWBox(MeasureIABase):
 		Num_shape = len(positions_shape_sample)  # number of halos in shape sample
 		LOS_ind = self.data["LOS"]  # eg 2 for z axis
 		not_LOS = np.array([0, 1, 2])[np.isin([0, 1, 2], LOS_ind, invert=True)]  # eg 0,1 for x&y
-		e = (1 - q ** 2) / (1 + q ** 2)  # size of ellipticity
+		if ellipticity == 'distortion':
+			e = (1 - q ** 2) / (1 + q ** 2)  # size of ellipticity
+		elif ellipticity == 'ellipticity':
+			e = (1 - q) / (1 + q)
+		else:
+			raise ValueError("Invalid value for ellipticity. Choose 'distortion' or 'ellipticity'.")
 		del q
 		R = sum(weight_shape * (1 - e ** 2 / 2.0)) / sum(weight_shape)
 		# R = 1 - np.mean(e ** 2) / 2.0  # responsitivity factor
@@ -298,21 +306,9 @@ class MeasureWBox(MeasureIABase):
 		Scross_D = np.array([[0.0] * self.num_bins_pi] * self.num_bins_r)
 		RR_g_plus = np.array([[0.0] * self.num_bins_pi] * self.num_bins_r)
 		RR_gg = np.array([[0.0] * self.num_bins_pi] * self.num_bins_r)
-		variance = np.array([[0.0] * self.num_bins_pi] * self.num_bins_r)
-		if tree_input != None:
-			indices_not_position, indices_shape = tree_input[0], tree_input[1]
-			Num_position -= len(indices_not_position)
-			Num_shape = len(indices_shape)
-			R = 1 - np.mean(e[indices_shape] ** 2) / 2.0
-			tree_file = open(f"{file_tree_path}/{dataset_name_tree}.pickle", 'rb')
-		if print_num:
-			print(
-				f"There are {Num_shape} galaxies in the shape sample and {Num_position} galaxies in the position sample.")
-		figname_dataset_name = dataset_name
-		if "/" in dataset_name:
-			figname_dataset_name = figname_dataset_name.replace("/", "_")
-		if "." in dataset_name:
-			figname_dataset_name = figname_dataset_name.replace(".", "p")
+
+		print(
+			f"There are {Num_shape} galaxies in the shape sample and {Num_position} galaxies in the position sample.")
 		pos_tree = KDTree(positions[:, not_LOS], boxsize=self.boxsize)
 		for i in np.arange(0, len(positions_shape_sample), 100):
 			i2 = min(len(positions_shape_sample), i + 100)
@@ -320,22 +316,10 @@ class MeasureWBox(MeasureIABase):
 			axis_direction_i = axis_direction[i:i2]
 			e_i = e[i:i2]
 			weight_shape_i = weight_shape[i:i2]
-			if tree_input != None:
-				ind_rbin = pickle.load(tree_file)
-				indices_shape_i = indices_shape[(indices_shape >= i) * (indices_shape < i2)] - i
-				ind_rbin_i = self.setdiff_omit(ind_rbin, indices_not_position, indices_shape_i)
-				positions_shape_sample_i = positions_shape_sample_i[indices_shape_i]
-				axis_direction_i = axis_direction_i[indices_shape_i]
-				e_i = e_i[indices_shape_i]
-				weight_shape_i = weight_shape_i[indices_shape_i]
-			else:
-				shape_tree = KDTree(positions_shape_sample_i[:, not_LOS], boxsize=self.boxsize)
-				ind_min_i = shape_tree.query_ball_tree(pos_tree, self.r_min)
-				ind_max_i = shape_tree.query_ball_tree(pos_tree, self.r_max)
-				ind_rbin_i = self.setdiff2D(ind_max_i, ind_min_i)
-				if save_tree:
-					with open(f"{file_tree_path}/w_{self.simname}_tree_{figname_dataset_name}.pickle", 'ab') as handle:
-						pickle.dump(ind_rbin_i, handle, protocol=pickle.HIGHEST_PROTOCOL)
+			shape_tree = KDTree(positions_shape_sample_i[:, not_LOS], boxsize=self.boxsize)
+			ind_min_i = shape_tree.query_ball_tree(pos_tree, self.r_min)
+			ind_max_i = shape_tree.query_ball_tree(pos_tree, self.r_max)
+			ind_rbin_i = self.setdiff2D(ind_max_i, ind_min_i)
 			for n in np.arange(0, len(positions_shape_sample_i)):  # CHANGE2: loop now over shapes, not positions
 				if len(ind_rbin_i[n]) > 0:
 					# for Splus_D (calculate ellipticities around position sample)
@@ -369,6 +353,10 @@ class MeasureWBox(MeasureIABase):
 						LOS[mask] / sub_box_len_pi - self.pi_bins[0] / sub_box_len_pi
 					)  # need length of LOS, so only positive values
 					ind_pi = np.array(ind_pi, dtype=int)
+					if np.any(ind_pi == self.num_bins_pi):
+						ind_pi[ind_pi >= self.num_bins_pi] -= 1
+					if np.any(ind_r == self.num_bins_r):
+						ind_r[ind_r >= self.num_bins_r] -= 1
 					np.add.at(Splus_D, (ind_r, ind_pi),
 							  (weight[ind_rbin_i[n]][mask] * weight_shape_i[n] * e_plus[mask]) / (2 * R))
 					np.add.at(Scross_D, (ind_r, ind_pi),
@@ -376,8 +364,6 @@ class MeasureWBox(MeasureIABase):
 					del e_plus, e_cross, separation_len
 					np.add.at(DD, (ind_r, ind_pi), weight[ind_rbin_i[n]][mask] * weight_shape_i[n])
 
-		if tree_input != None:
-			tree_file.close()
 		# if Num_position == Num_shape:
 		# 	corrtype = "auto"
 		# 	DD = DD / 2.0  # auto correlation, all pairs are double
@@ -393,7 +379,6 @@ class MeasureWBox(MeasureIABase):
 					Num_position, Num_shape)
 		correlation = Splus_D / RR_g_plus  # (Splus_D - Splus_R) / RR_g_plus
 		xi_g_cross = Scross_D / RR_g_plus  # (Scross_D - Scross_R) / RR_g_plus
-		sigsq = variance / RR_g_plus ** 2
 		dsep = (self.r_bins[1:] - self.r_bins[:-1]) / 2.0
 		separation_bins = self.r_bins[:-1] + abs(dsep)  # middle of bins
 		dpi = (self.pi_bins[1:] - self.pi_bins[:-1]) / 2.0
@@ -405,21 +390,18 @@ class MeasureWBox(MeasureIABase):
 			write_dataset_hdf5(group, dataset_name, data=correlation)
 			write_dataset_hdf5(group, dataset_name + "_SplusD", data=Splus_D)
 			write_dataset_hdf5(group, dataset_name + "_RR_g_plus", data=RR_g_plus)
-			write_dataset_hdf5(group, dataset_name + "_sigmasq", data=sigsq)
 			write_dataset_hdf5(group, dataset_name + "_rp", data=separation_bins)
 			write_dataset_hdf5(group, dataset_name + "_pi", data=pi_bins)
 			group = create_group_hdf5(output_file, f"{self.snap_group}/w/xi_g_cross/{jk_group_name}")
 			write_dataset_hdf5(group, dataset_name + "_ScrossD", data=Scross_D)
 			write_dataset_hdf5(group, dataset_name, data=xi_g_cross)
 			write_dataset_hdf5(group, dataset_name + "_RR_g_cross", data=RR_g_plus)
-			write_dataset_hdf5(group, dataset_name + "_sigmasq", data=sigsq)
 			write_dataset_hdf5(group, dataset_name + "_rp", data=separation_bins)
 			write_dataset_hdf5(group, dataset_name + "_pi", data=pi_bins)
 			group = create_group_hdf5(output_file, f"{self.snap_group}/w/xi_gg/{jk_group_name}")
 			write_dataset_hdf5(group, dataset_name, data=(DD / RR_gg) - 1)
 			write_dataset_hdf5(group, dataset_name + "_DD", data=DD)
 			write_dataset_hdf5(group, dataset_name + "_RR_gg", data=RR_gg)
-			write_dataset_hdf5(group, dataset_name + "_sigmasq", data=sigsq)
 			write_dataset_hdf5(group, dataset_name + "_rp", data=separation_bins)
 			write_dataset_hdf5(group, dataset_name + "_pi", data=pi_bins)
 			output_file.close()
@@ -427,123 +409,133 @@ class MeasureWBox(MeasureIABase):
 		else:
 			return correlation, (DD / RR_gg) - 1, separation_bins, pi_bins, Splus_D, DD, RR_g_plus
 
-	def _measure_xi_rp_pi_sims_batch(self, indices):
-		"""
+	def _measure_xi_rp_pi_box_batch(self, i):
+		r"""Measures components of $\xi_{gg}$ and $\xi_{g+}$ in (rp,pi) bins including jackknife realisations for a batch
+		of indices from i to i+chunk_size. Support function for _measure_xi_rp_pi_box_jk_multiprocessing().
 
 		Parameters
 		----------
-		indices :
-		    
+		i: int
+			Start index of the batch.
 
 		Returns
 		-------
-
+		ndarrays
+			S+D, SxD, DD, DD_jk, S+D_jk where the _jk versions store the necessary information of DD of S+D for
+			each jackknife realisation.
 		"""
+		if i + self.chunk_size > self.Num_shape_masked:
+			i2 = self.Num_shape_masked
+		else:
+			i2 = i + self.chunk_size
+
 		DD = np.array([[0.0] * self.num_bins_pi] * self.num_bins_r)
 		Splus_D = np.array([[0.0] * self.num_bins_pi] * self.num_bins_r)
 		Scross_D = np.array([[0.0] * self.num_bins_pi] * self.num_bins_r)
-		variance = np.array([[0.0] * self.num_bins_pi] * self.num_bins_r)
 
-		for j in np.arange(0, len(indices), 100):
-			i = indices[j]
-			i2 = min(indices[-1], i + 100)
-			positions_shape_sample_i = self.positions_shape_sample[i:i2]
-			axis_direction_i = self.axis_direction[i:i2]
-			e_i = self.e[i:i2]
-			weight_shape_i = self.weight_shape[i:i2]
+		positions_shape_sample_i = self.temp_data_obj.read_cat("positions_shape_sample", [i, i2])
+		axis_direction_i = self.temp_data_obj.read_cat("axis_direction", [i, i2])
+		weight_shape_i = self.temp_data_obj.read_cat("weight_shape", [i, i2])
+		positions = self.temp_data_obj.read_cat("positions")
+		weight = self.temp_data_obj.read_cat("weight")
+		e_i = self.e[i:i2]
 
-			shape_tree = KDTree(positions_shape_sample_i[:, self.not_LOS], boxsize=self.boxsize)
-			ind_min_i = shape_tree.query_ball_tree(self.pos_tree, self.r_min)
-			ind_max_i = shape_tree.query_ball_tree(self.pos_tree, self.r_max)
-			ind_rbin_i = self.setdiff2D(ind_max_i, ind_min_i)
-			for n in np.arange(0, len(positions_shape_sample_i)):  # CHANGE2: loop now over shapes, not positions
-				if len(ind_rbin_i[n]) > 0:
-					# for Splus_D (calculate ellipticities around position sample)
-					separation = positions_shape_sample_i[n] - self.positions[ind_rbin_i[n]]  # CHANGE1 & CHANGE2
-					if self.periodicity:
-						separation[separation > self.L_0p5] -= self.boxsize  # account for periodicity of box
-						separation[separation < -self.L_0p5] += self.boxsize
-					projected_sep = separation[:, self.not_LOS]
-					LOS = separation[:, self.LOS_ind]
-					separation_len = np.sqrt(np.sum(projected_sep ** 2, axis=1))
-					with np.errstate(invalid='ignore'):
-						separation_dir = (projected_sep.transpose() / separation_len).transpose()  # normalisation of rp
-						del projected_sep, separation
-						phi = np.arccos(
-							separation_dir[:, 0] * axis_direction_i[n, 0] + separation_dir[:, 1] * axis_direction_i[
-								n, 1])  # CHANGE2
-					e_plus, e_cross = self.get_ellipticity(e_i[n], phi)  # CHANGE2
-					del phi, separation_dir
-					e_plus[np.isnan(e_plus)] = 0.0
-					e_cross[np.isnan(e_cross)] = 0.0
+		shape_tree = KDTree(positions_shape_sample_i[:, self.not_LOS], boxsize=self.boxsize)
+		ind_min_i = shape_tree.query_ball_tree(self.pos_tree, self.r_min)
+		ind_max_i = shape_tree.query_ball_tree(self.pos_tree, self.r_max)
+		ind_rbin_i = self.setdiff2D(ind_max_i, ind_min_i)
+		for n in np.arange(0, len(positions_shape_sample_i)):  # CHANGE2: loop now over shapes, not positions
+			if len(ind_rbin_i[n]) > 0:
+				# for Splus_D (calculate ellipticities around position sample)
+				separation = positions_shape_sample_i[n] - positions[ind_rbin_i[n]]  # CHANGE1 & CHANGE2
+				if self.periodicity:
+					separation[separation > self.L_0p5] -= self.boxsize  # account for periodicity of box
+					separation[separation < -self.L_0p5] += self.boxsize
+				projected_sep = separation[:, self.not_LOS]
+				LOS = separation[:, self.LOS_ind]
+				separation_len = np.sqrt(np.sum(projected_sep ** 2, axis=1))
+				with np.errstate(invalid='ignore'):
+					separation_dir = (projected_sep.transpose() / separation_len).transpose()  # normalisation of rp
+					del projected_sep, separation
+					phi = np.arccos(
+						separation_dir[:, 0] * axis_direction_i[n, 0] + separation_dir[:, 1] * axis_direction_i[
+							n, 1])  # CHANGE2
+				e_plus, e_cross = self.get_ellipticity(e_i[n], phi)  # CHANGE2
+				del phi, separation_dir
+				e_plus[np.isnan(e_plus)] = 0.0
+				e_cross[np.isnan(e_cross)] = 0.0
 
-					# get the indices for the binning
-					mask = (separation_len >= self.r_bins[0]) * (separation_len < self.r_bins[-1]) * (
-							LOS >= self.pi_bins[0]) * (LOS < self.pi_bins[-1])
-					ind_r = np.floor(
-						np.log10(separation_len[mask]) / self.sub_box_len_logrp - np.log10(
-							self.r_bins[0]) / self.sub_box_len_logrp
-					)
-					ind_r = np.array(ind_r, dtype=int)
-					ind_pi = np.floor(
-						LOS[mask] / self.sub_box_len_pi - self.pi_bins[0] / self.sub_box_len_pi
-					)  # need length of LOS, so only positive values
-					ind_pi = np.array(ind_pi, dtype=int)
-					np.add.at(Splus_D, (ind_r, ind_pi),
-							  (self.weight[ind_rbin_i[n]][mask] * weight_shape_i[n] * e_plus[mask]) / (2 * self.R))
-					np.add.at(Scross_D, (ind_r, ind_pi),
-							  (self.weight[ind_rbin_i[n]][mask] * weight_shape_i[n] * e_cross[mask]) / (2 * self.R))
-					del e_plus, e_cross, separation_len
-					# np.add.at(variance, (ind_r, ind_pi), (e_plus[mask] / (2 * R)) ** 2)
-					np.add.at(DD, (ind_r, ind_pi), self.weight[ind_rbin_i[n]][mask] * weight_shape_i[n])
+				# get the indices for the binning
+				mask = (separation_len >= self.r_bins[0]) * (separation_len < self.r_bins[-1]) * (
+						LOS >= self.pi_bins[0]) * (LOS < self.pi_bins[-1])
+				ind_r = np.floor(
+					np.log10(separation_len[mask]) / self.sub_box_len_logrp - np.log10(
+						self.r_bins[0]) / self.sub_box_len_logrp
+				)
+				ind_r = np.array(ind_r, dtype=int)
+				ind_pi = np.floor(
+					LOS[mask] / self.sub_box_len_pi - self.pi_bins[0] / self.sub_box_len_pi
+				)  # need length of LOS, so only positive values
+				ind_pi = np.array(ind_pi, dtype=int)
+				if np.any(ind_pi == self.num_bins_pi):
+					ind_pi[ind_pi >= self.num_bins_pi] -= 1
+				if np.any(ind_r == self.num_bins_r):
+					ind_r[ind_r >= self.num_bins_r] -= 1
+				np.add.at(Splus_D, (ind_r, ind_pi),
+						  (weight[ind_rbin_i[n]][mask] * weight_shape_i[n] * e_plus[mask]) / (2 * self.R))
+				np.add.at(Scross_D, (ind_r, ind_pi),
+						  (weight[ind_rbin_i[n]][mask] * weight_shape_i[n] * e_cross[mask]) / (2 * self.R))
+				del separation_len, e_cross, e_plus
+				np.add.at(DD, (ind_r, ind_pi), weight[ind_rbin_i[n]][mask] * weight_shape_i[n])
 
-		return Splus_D, Scross_D, DD, variance
+		return Splus_D, Scross_D, DD
 
-	def _measure_xi_rp_pi_sims_multiprocessing(self, num_nodes=9, masks=None,
-											   dataset_name="All_galaxies", return_output=False,
-											   print_num=True, jk_group_name=""):
-		"""Measures the projected correlation function (xi_g_plus, xi_gg) for given coordinates of the position and shape sample
-		(Position, Position_shape_sample), the projected axis direction (Axis_Direction), the ratio between projected
-		axes, q=b/a (q) and the index of the direction of the line of sight (LOS=2 for z axis).
-		Positions are assumed to be given in cMpc/h.
+	def _measure_xi_rp_pi_box_multiprocessing(self, dataset_name, temp_file_path, masks=None,
+											  return_output=False, jk_group_name="", num_nodes=1, chunk_size=1000,
+											  ellipticity='distortion'):
+		r"""Measures the projected correlation functions, $\xi_{gg}$ and $\xi_{g+}$, in (rp, pi) bins for an object
+		created with MeasureIABox. Uses >1 CPU. Uses KDTree for speedup.
 
 		Parameters
 		----------
-		masks :
-		    the masks for the data to select only part of the data (Default value = None)
-		dataset_name :
-		    the dataset name given in the hdf5 file. (Default value = "All_galaxies")
-		return_output :
-		    Output is returned if True, saved to file if False. (Default value = False)
-		num_nodes :
-		     (Default value = 9)
-		print_num :
-		     (Default value = True)
-		jk_group_name :
-		     (Default value = "")
+		dataset_name : str
+			Name of the dataset in the output file.
+		temp_file_path : str or NoneType, optional
+			Path to where the data is temporarily stored [file name generated automatically].
+		num_nodes : int, optional
+			Number of CPUs used in the multiprocessing. Default is 1.
+		masks : dict or NoneType, optional
+			Dictionary with masks for the data to select only part of the data. Uses same keywords as data dictionary.
+			Default value = None.
+		return_output : bool, optional
+			If True, the output will be returned instead of written to a file. Default value is False.
+		jk_group_name : str, optional
+			Group in output file (hdf5) where jackknife realisations are stored. Default value is "".
+		ellipticity : str, optional
+			Definition of ellipticity. Choose from 'distortion', defined as (1-q^2)/(1+q^2), or 'ellipticity', defined
+			 as (1-q)/(1+q). Default is 'distortion'.
 
 		Returns
 		-------
-		type
-		    xi_g_plus, xi_gg, separation_bins, pi_bins if no output file is specified
-
+		ndarrays
+			$\xi_{gg}$ and $\xi_{g+}$, r_p bins, pi bins, S+D, DD, RR (if no output file is specified)
 		"""
 
 		if masks == None:
-			self.positions = self.data["Position"]
-			self.positions_shape_sample = self.data["Position_shape_sample"]
+			positions = self.data["Position"]
+			positions_shape_sample = self.data["Position_shape_sample"]
 			axis_direction_v = self.data["Axis_Direction"]
 			axis_direction_len = np.sqrt(np.sum(axis_direction_v ** 2, axis=1))
-			self.axis_direction = (axis_direction_v.transpose() / axis_direction_len).transpose()
+			axis_direction = (axis_direction_v.transpose() / axis_direction_len).transpose()
 			q = self.data["q"]
-			self.weight = self.data["weight"]
-			self.weight_shape = self.data["weight_shape_sample"]
+			weight = self.data["weight"]
+			weight_shape = self.data["weight_shape_sample"]
 		else:
-			self.positions = self.data["Position"][masks["Position"]]
-			self.positions_shape_sample = self.data["Position_shape_sample"][masks["Position_shape_sample"]]
+			positions = self.data["Position"][masks["Position"]]
+			positions_shape_sample = self.data["Position_shape_sample"][masks["Position_shape_sample"]]
 			axis_direction_v = self.data["Axis_Direction"][masks["Axis_Direction"]]
 			axis_direction_len = np.sqrt(np.sum(axis_direction_v ** 2, axis=1))
-			self.axis_direction = (axis_direction_v.transpose() / axis_direction_len).transpose()
+			axis_direction = (axis_direction_v.transpose() / axis_direction_len).transpose()
 			q = self.data["q"][masks["q"]]
 			try:
 				weight_mask = masks["weight"]
@@ -555,19 +547,39 @@ class MeasureWBox(MeasureIABase):
 			except:
 				masks["weight_shape_sample"] = np.ones(self.Num_shape, dtype=bool)
 				masks["weight_shape_sample"][sum(masks["Position_shape_sample"]):self.Num_shape] = 0
-			self.weight = self.data["weight"][masks["weight"]]
-			self.weight_shape = self.data["weight_shape_sample"][masks["weight_shape_sample"]]
+			weight = self.data["weight"][masks["weight"]]
+			weight_shape = self.data["weight_shape_sample"][masks["weight_shape_sample"]]
 		# masking changes the number of galaxies
-		Num_position = len(self.positions)  # number of halos in position sample
-		Num_shape = len(self.positions_shape_sample)  # number of halos in shape sample
-		if print_num:
-			print(
-				f"There are {Num_shape} galaxies in the shape sample and {Num_position} galaxies in the position sample.")
+		self.Num_position_masked = len(positions)
+		self.Num_shape_masked = len(positions_shape_sample)
+		print(
+			f"There are {self.Num_shape_masked} galaxies in the shape sample and {self.Num_position_masked} galaxies in the position sample.")
+
+		# create temp hdf5 from which data can be read. del self.data, but save it in this method to reduce RAM
+		figname_dataset_name = dataset_name
+		if "/" in dataset_name:
+			figname_dataset_name = figname_dataset_name.replace("/", "_")
+		if "." in dataset_name:
+			figname_dataset_name = figname_dataset_name.replace(".", "p")
+		file_temp = h5py.File(f"{temp_file_path}/w_{self.simname}_temp_data_{figname_dataset_name}.hdf5", "w")
+		write_dataset_hdf5(file_temp, "positions", positions)
+		write_dataset_hdf5(file_temp, "weight", weight)
+		write_dataset_hdf5(file_temp, "weight_shape", weight_shape)
+		write_dataset_hdf5(file_temp, "positions_shape_sample", positions_shape_sample)
+		write_dataset_hdf5(file_temp, "axis_direction", axis_direction)
+		file_temp.close()
+		self.temp_data_obj = ReadData(self.simname, f"w_{self.simname}_temp_data_{figname_dataset_name}", None,
+									  data_path=temp_file_path)
 
 		self.LOS_ind = self.data["LOS"]  # eg 2 for z axis
 		self.not_LOS = np.array([0, 1, 2])[np.isin([0, 1, 2], self.LOS_ind, invert=True)]  # eg 0,1 for x&y
-		self.e = (1 - q ** 2) / (1 + q ** 2)  # size of ellipticity
-		self.R = sum(self.weight_shape * (1 - self.e ** 2 / 2.0)) / sum(self.weight_shape)
+		if ellipticity == 'distortion':
+			self.e = (1 - q ** 2) / (1 + q ** 2)  # size of ellipticity
+		elif ellipticity == 'ellipticity':
+			self.e = (1 - q) / (1 + q)
+		else:
+			raise ValueError("Invalid value for ellipticity. Choose 'distortion' or 'ellipticity'.")
+		self.R = sum(weight_shape * (1 - self.e ** 2 / 2.0)) / sum(weight_shape)
 		# self.R = 1 - np.mean(self.e ** 2) / 2.0  # responsitivity factor
 		L3 = self.boxsize ** 3  # box volume
 		self.sub_box_len_logrp = (np.log10(self.r_max) - np.log10(self.r_min)) / self.num_bins_r
@@ -577,20 +589,25 @@ class MeasureWBox(MeasureIABase):
 		Scross_D = np.array([[0.0] * self.num_bins_pi] * self.num_bins_r)
 		RR_g_plus = np.array([[0.0] * self.num_bins_pi] * self.num_bins_r)
 		RR_gg = np.array([[0.0] * self.num_bins_pi] * self.num_bins_r)
-		variance = np.array([[0.0] * self.num_bins_pi] * self.num_bins_r)
 
-		self.pos_tree = KDTree(self.positions[:, self.not_LOS], boxsize=self.boxsize)
+		data_temp = self.data  # make sure data is not sent to every CPU
+		self.data = None
 
-		self.multiproc_chuncks = np.array_split(np.arange(len(self.positions_shape_sample)), num_nodes)
-		result = ProcessingPool(nodes=num_nodes).map(
-			self._measure_xi_rp_pi_sims_batch,
-			self.multiproc_chuncks,
-		)
-		for i in np.arange(num_nodes):
+		self.pos_tree = KDTree(positions[:, self.not_LOS], boxsize=self.boxsize)
+		indices = np.arange(0, len(positions_shape_sample), chunk_size)
+		self.chunk_size = chunk_size
+		with Pool(num_nodes) as p:
+			result = p.map(self._measure_xi_rp_pi_box_batch, indices)
+		os.remove(
+			f"{temp_file_path}/w_{self.simname}_temp_data_{figname_dataset_name}.hdf5")
+
+		self.data = data_temp
+		del data_temp
+
+		for i in np.arange(len(result)):
 			Splus_D += result[i][0]
 			Scross_D += result[i][1]
 			DD += result[i][2]
-			variance += result[i][3]
 
 		# if Num_position == Num_shape:
 		# 	corrtype = "auto"
@@ -602,13 +619,12 @@ class MeasureWBox(MeasureIABase):
 			for p in np.arange(0, self.num_bins_pi):
 				RR_g_plus[i, p] = self.get_random_pairs(
 					self.r_bins[i + 1], self.r_bins[i], self.pi_bins[p + 1], self.pi_bins[p], L3, "cross",
-					Num_position, Num_shape)
+					self.Num_position_masked, self.Num_shape_masked)
 				RR_gg[i, p] = self.get_random_pairs(
 					self.r_bins[i + 1], self.r_bins[i], self.pi_bins[p + 1], self.pi_bins[p], L3, corrtype,
-					Num_position, Num_shape)
+					self.Num_position_masked, self.Num_shape_masked)
 		correlation = Splus_D / RR_g_plus  # (Splus_D - Splus_R) / RR_g_plus
 		xi_g_cross = Scross_D / RR_g_plus  # (Scross_D - Scross_R) / RR_g_plus
-		sigsq = variance / RR_g_plus ** 2
 		dsep = (self.r_bins[1:] - self.r_bins[:-1]) / 2.0
 		separation_bins = self.r_bins[:-1] + abs(dsep)  # middle of bins
 		dpi = (self.pi_bins[1:] - self.pi_bins[:-1]) / 2.0
@@ -620,21 +636,18 @@ class MeasureWBox(MeasureIABase):
 			write_dataset_hdf5(group, dataset_name, data=correlation)
 			write_dataset_hdf5(group, dataset_name + "_SplusD", data=Splus_D)
 			write_dataset_hdf5(group, dataset_name + "_RR_g_plus", data=RR_g_plus)
-			write_dataset_hdf5(group, dataset_name + "_sigmasq", data=sigsq)
 			write_dataset_hdf5(group, dataset_name + "_rp", data=separation_bins)
 			write_dataset_hdf5(group, dataset_name + "_pi", data=pi_bins)
 			group = create_group_hdf5(output_file, f"{self.snap_group}/w/xi_g_cross/{jk_group_name}")
 			write_dataset_hdf5(group, dataset_name + "_ScrossD", data=Scross_D)
 			write_dataset_hdf5(group, dataset_name, data=xi_g_cross)
 			write_dataset_hdf5(group, dataset_name + "_RR_g_cross", data=RR_g_plus)
-			write_dataset_hdf5(group, dataset_name + "_sigmasq", data=sigsq)
 			write_dataset_hdf5(group, dataset_name + "_rp", data=separation_bins)
 			write_dataset_hdf5(group, dataset_name + "_pi", data=pi_bins)
 			group = create_group_hdf5(output_file, f"{self.snap_group}/w/xi_gg/{jk_group_name}")
 			write_dataset_hdf5(group, dataset_name, data=(DD / RR_gg) - 1)
 			write_dataset_hdf5(group, dataset_name + "_DD", data=DD)
 			write_dataset_hdf5(group, dataset_name + "_RR_gg", data=RR_gg)
-			write_dataset_hdf5(group, dataset_name + "_sigmasq", data=sigsq)
 			write_dataset_hdf5(group, dataset_name + "_rp", data=separation_bins)
 			write_dataset_hdf5(group, dataset_name + "_pi", data=pi_bins)
 			output_file.close()
