@@ -5,9 +5,9 @@ def write_dataset_hdf5(group, name, data):
 	----------
 	group : <class 'h5py._hl.group.Group'>
 		link to group in datafile
-	name :
+	name : str
 		name of desired dataset
-	data :
+	data : ndarray
 		the data
 
 	"""
@@ -43,17 +43,16 @@ def create_group_hdf5(file, name):
 		list_groups = name.split('/')
 		groups = ''
 		for group in list_groups:
-			if groups == '':
-				groups += group
-			else:
-				groups += '/'
-				groups += group
-			try:
-				group_file = file[groups]
-			except:
-				file.create_group(groups)
-	if groups != name:
-		raise ValueError('Created groups do not match specified one.')
+			if group != '':
+				if groups == '':
+					groups += group
+				else:
+					groups += '/'
+					groups += group
+				try:
+					group_file = file[groups]
+				except:
+					file.create_group(groups)
 	return file[name]
 
 
